/* st-accel
 * 
 * A driver for the ST LIS3L02DQ 3-axis accelerometer on Lama's sensor board
 *
 * Author: Karl Koscher
 * Version: 0.4
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/kernel.h> // For debugging only
#include <asm/signal.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/hardware.h>
#include <asm/arch/ssp.h>
#include <asm-arm/arch-pxa/pxa-regs.h> // This include file lets us access
									   // memory-mapped I/O registers

#include "st-accel.h"

// These macros include some documentation in the final module file,
// such as a description of the module, the author, and version.
// These macros are optional, but you're encouraged to use them

MODULE_AUTHOR("Karl Koscher");
MODULE_DESCRIPTION("A driver for the ST LIS3L02DQ 3-axis accelerometer on Lama's sensor board");
MODULE_VERSION("0.5");

// This macro specifies which license this module is released under.
// This macro is optional, but if you do not specify "GPL", "GPL v2",
// "GPL and additional rights", "Dual BSD/GPL", or "Dual MPL/GPL",
// the kernel will complain that it has been tainted with non-free code.
// Additionally, you may not be able to use certain kernel functions if
// your module is not licensed under the GPL.

// MODULE_LICENSE("GPL");

dev_t devId; // Contains the major and minor device numbers
struct cdev *cdev; // A kernel character device struct
struct ssp_dev accel_ssp;

int port = 2;

#define DATABUF_SIZE_POW2	14 // 16 KB
#define DATABUF_SIZE		(1 << DATABUF_SIZE_POW2)
#define DATABUF_SIZE_MASK	(DATABUF_SIZE - 1)

unsigned char dataBuf[DATABUF_SIZE];
volatile unsigned int curPtr = 0;
//unsigned int bufSize = 0; // FIXME: Remember when we've overflown our buffer

DECLARE_WAIT_QUEUE_HEAD(waiting_for_data_queue);

ssize_t accel_read(struct file *, char __user *, size_t, loff_t *);
int accel_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static void __exit unload_function(void);

// This struct contains pointers to all supported file operations
struct file_operations accel_fops = {
	.owner = THIS_MODULE,
	.read = accel_read,
	.ioctl = accel_ioctl,
};

// Allow setting the port as an initialization parameter
module_param(port, int, 0);

unsigned char accel_read_reg(int reg)
{
	ssp_write_word(&accel_ssp, (reg | 0x80) << 8);
	return ssp_read_word(&accel_ssp) & 0xFF;
}

void accel_write_reg(int reg, unsigned char data)
{
	ssp_write_word(&accel_ssp, (reg << 8) | data);
	ssp_read_word(&accel_ssp);
}

ssize_t accel_read(struct file *filp, char __user *userData, size_t count, loff_t * offp)
{
	int returned_count = 0;
	int cptr;

	*offp &= DATABUF_SIZE_MASK; // Be really paranoid about the offset being a legal value

	if (wait_event_interruptible(waiting_for_data_queue, curPtr != *offp))
		return -ERESTARTSYS;
	
	// Get a snapshot of curPtr just to be safe
	cptr = curPtr;
	
	if (*offp > cptr) {
		if (count > (DATABUF_SIZE - *offp))
			returned_count = (DATABUF_SIZE - *offp);
		else
			returned_count = count;

		if (copy_to_user(userData, dataBuf + *offp, returned_count))
			return -EFAULT;

		count -= returned_count;
		*offp = 0;
	}

	if (count) {
		if (count > (cptr - *offp))
			count = (cptr - *offp);
	
		if (copy_to_user(userData + returned_count, dataBuf + *offp, count))
			return -EFAULT;
	
		*offp += count;
		returned_count += count;
	}

	return returned_count;
}

int accel_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	unsigned char data;

	switch (cmd) {
	
	/**
	TODO:
	Fill in appropriate ioctl calls here

	**/

		default:
			return -ENOTTY; // It seems like this should be -EINVAL, but 
							// apparently -ENOTTY also means "so such ioctl"
	}

	return 0;
}

// Write data into the circular buffer in a safe way
// Note that only one instance of this function can be active
// at a time, and writing the data and current pointer are atomic operations,
// we can safely update the buffer without any locks

inline void write_data_into_buffer(unsigned char data)
{
	dataBuf[curPtr] = data;
	curPtr = (curPtr + 1) & DATABUF_SIZE_MASK;
}

irqreturn_t accel_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/**

	TODO:

	Fill in appropriate irq_handler code here


	**/
}

static int __init init_function(void)
{
	int result;

	// Ensure that the port parameter is valid
	if (port < 1 || port > 3)
		return -ENODEV;

	// Allocate a major device number for this driver
	result = alloc_chrdev_region(&devId, 0, 1, "st-accel");
	if (result < 0)
		return result;

	// Allocate a character device
	cdev = cdev_alloc();
	if (!cdev) {
		printk(KERN_WARNING "st-accel: unable to allocate character device\n");
		unregister_chrdev_region(devId, 1);
		return -EAGAIN;
	}

	// Set the owner and file operations of this new character device
	cdev->owner = THIS_MODULE;
	cdev->ops = &accel_fops;

	// Register the character device
	// Note at this point the device is live!
	result = cdev_add(cdev, devId, 1);
	if (result < 0) {
		printk(KERN_WARNING "st-accel: unable to add character device\n");
		cdev_del(cdev);
		unregister_chrdev_region(devId, 1);
		return result;
	}

	result = request_irq(IRQ_GPIO(96), accel_irq_handler, SA_SHIRQ, "st-accel", &devId);
	if (result < 0) {
		printk(KERN_WARNING "st-accel: unable to request IRQ\n");
		cdev_del(cdev);
		unregister_chrdev_region(devId, 1);
		return result;
	}

	// Initialize the GPIO pins
	// Note that we need to ensure that no one else is messing with these
	// registers while we adjust them, so we disable interrupts

	local_irq_disable();
						
	GPDR1 |= 0x00000040; // GPIO36 (SSP TXD2) output
	
	GAFR0_L |= 0x00800000;
	GAFR1_L |= 0x00002A00; // Set SSPCLK, SSPFRM, SSPTXD, and SSPRXD functions
	
	// GPIO is shared between the temperature sensor and the RDY_INT signal
	GAFR3_L &= ~(0x00000003); // Set GPIO 96 as a GPIO pin
	GPDR3 &= ~(0x00000001); // Set GPIO 96 as an input
	GRER3 |= 0x00000001; // Interrupt on the rising edge of GPIO 96
	GFER3 &= ~0x00000001;
	

	ICMR |= 0x00000400;	// Enable GPIO interrupts
	local_irq_enable();

	// Initialize the SSP
	result = ssp_init(&accel_ssp, port);
	if (result < 0) {
		printk(KERN_WARNING "st-accel: unable to initialize SSP\n");
		cdev_del(cdev);
		unregister_chrdev_region(devId, 1);
		return result;
	}

	// SSCR0_Motorola is SPI. Since I don't know what VDD_io is set at, 
	// we make the safe assumption that it can only run up to 4 MHz
	ssp_config(&accel_ssp, SSCR0_Motorola | SSCR0_DataSize(16),
		SSCR1_SPO | SSCR1_SPH, 0, 3 /* 3.25 MHz */ );
	ssp_enable(&accel_ssp);

	// Enable data ready signal
	// TODO: Set BDU as well?
	accel_write_reg(CTRL_REG2, 0x05);

	return 0; // SUCCESS
}

static void __exit unload_function(void)
{
	// Disable interrupts
	accel_write_reg(CTRL_REG2, 0x00);
	free_irq(IRQ_GPIO(96), &devId);

	// Power-down the accelerometer, and reset defaults
	accel_write_reg(CTRL_REG1, 0x07);

	// Stop the SSP
	ssp_exit(&accel_ssp);
	// Free the character device
	cdev_del(cdev);
	// Unregister the major device number
	unregister_chrdev_region(devId, 1);
}

// These two macros let the compiler and kernel know which functions 
// should be called when loading and unloading the kernel.

module_init(init_function);
module_exit(unload_function);
